using System;
using System.Runtime.InteropServices;
using System.Text;

namespace Feynman
{
    class FeynmanClient
    {

        public delegate int DCWCallback(int notUsed, ref t_feynman_ecm ecm);

        #region DLL imports

        [DllImport("libfeynman")]
        private static extern void feynman_compile_time();

        [DllImport("libfeynman")]
        private static unsafe extern IntPtr feynman_add_remote_server(IntPtr server_db);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_remote_server_username(IntPtr server, [MarshalAs(UnmanagedType.LPStr)] string username);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_remote_server_password(IntPtr server, [MarshalAs(UnmanagedType.LPStr)] string password);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_remote_server_hostname(IntPtr server, [MarshalAs(UnmanagedType.LPStr)] string hostname);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_remote_server_port(IntPtr server, int port);

        [DllImport("libfeynman")]
        public static unsafe extern void feynman_set_remote_server_dcw_callback(IntPtr server, DCWCallback del);

        [DllImport("libfeynman")]
        private static unsafe extern int feynman_init_remote_server(IntPtr server);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_send_server_ecm(IntPtr server, ushort session, ushort caid, int provid, ushort sid, byte[] ecm_data, int len);

        [DllImport("libfeynman")]
        private static unsafe extern int feynman_dcw_dummy_callback(IntPtr server, ushort session, byte[] dwc, int timeout_ms);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_send_server_emm(IntPtr server, ushort caid, int provid, int sa, int ua, byte[] emm_data, int len);

        #endregion

        private bool isReady;
        private int ecmTimeout;
        private DCWCallback dcwCallback;
        private IntPtr pServer;
        private ushort session;

        public unsafe FeynmanClient(string username, string password, string hostname, int port, DCWCallback dcw_del)
        {
            this.isReady = false;
            this.session = 0;
            this.ecmTimeout = 5000;

            feynman_compile_time();

            this.pServer = feynman_add_remote_server(IntPtr.Zero);

            feynman_set_remote_server_username(this.pServer, username);
            feynman_set_remote_server_password(this.pServer, password);
            feynman_set_remote_server_hostname(this.pServer, hostname);
            feynman_set_remote_server_port(this.pServer, port);

            this.dcwCallback = new DCWCallback(dcw_del);
            feynman_set_remote_server_dcw_callback(this.pServer, this.dcwCallback);

            if(feynman_init_remote_server(this.pServer) > 0)
            {
                this.isReady = true;
            }
        }

        public unsafe int sendECM(ushort caid, int provid, ushort sid, byte[] ecm, int len, byte[] dcw)
        {
            this.session++;
            feynman_send_server_ecm(this.pServer, this.session, caid, provid, sid, ecm, len);
            if(this.dcwCallback == null)
            {
                return feynman_dcw_dummy_callback(this.pServer, this.session, dcw, this.ecmTimeout);
            }
            else
            {
                return 0;
            }
        }

        public unsafe void sendEMM(ushort caid, int provid, int sa, int ua, byte[] emm, int len)
        {
            feynman_send_server_emm(this.pServer, caid, provid, sa, ua, emm, len);
        }

        public void setTimeout(int timeout)
        {
            this.ecmTimeout = timeout;
        }

        public bool isServerReady()
        {
            if(this.isReady)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FeynmanServer
    {
        public delegate int ECMCallback(int notUsed, ref t_feynman_ecm ecm);
        public delegate int EMMCallback(int notUsed, ref t_feynman_emm emm);

        #region DLL imports

        [DllImport("libfeynman")]
        private static extern void feynman_compile_time();

        [DllImport("libfeynman")]
        private static unsafe extern IntPtr feynman_add_local_server(IntPtr server_db);

        [DllImport("libfeynman")]
        private static unsafe extern IntPtr feynman_add_client(IntPtr server);

        [DllImport("libfeynman")]
        private static unsafe extern IntPtr feynman_add_local_card(IntPtr server, int card_id);

        [DllImport("libfeynman")]
        private static unsafe extern IntPtr feynman_add_sid(IntPtr card, ushort new_sid, char mode);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_local_server_port(IntPtr server, int port);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_local_server_ecm_callback(IntPtr server, ECMCallback del);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_local_server_emm_callback(IntPtr server, EMMCallback del);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_client_username(IntPtr client, [MarshalAs(UnmanagedType.LPStr)] string username);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_client_password(IntPtr client, [MarshalAs(UnmanagedType.LPStr)] string password);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_client_access_mode(IntPtr client, int access_mode);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_local_card_caid(IntPtr card, ushort caid);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_local_card_provid(IntPtr card, int provid);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_local_card_sa(IntPtr card, int sa);

        [DllImport("libfeynman")]
        private static unsafe extern void feynman_set_local_card_ua(IntPtr card, int ua);

        [DllImport("libfeynman")]
        private static unsafe extern int feynman_start_local_server(IntPtr server);

        #endregion

        private IntPtr pServer;
        private int cardCount;
        private int userCount;
        private ECMCallback ecmCallback;
        private EMMCallback emmCallback;

        public unsafe FeynmanServer(int port, ECMCallback ecm_del, EMMCallback emm_del)
        {
            feynman_compile_time();
            this.pServer = feynman_add_local_server(IntPtr.Zero);

            this.ecmCallback = new ECMCallback(ecm_del);
            this.emmCallback = new EMMCallback(emm_del);

            feynman_set_local_server_ecm_callback(this.pServer, this.ecmCallback);
            feynman_set_local_server_emm_callback(this.pServer, this.emmCallback);
            feynman_set_local_server_port(this.pServer, port);

            this.cardCount = 0;
            this.userCount = 0;
        }

        public unsafe int addUser(string username, string password, int access_mode)
        {
            this.userCount++;
            IntPtr pClient = feynman_add_client(this.pServer);
            feynman_set_client_username(pClient, username);
            feynman_set_client_password(pClient, password);
            feynman_set_client_access_mode(pClient, access_mode);

            return this.userCount;
        }

        public unsafe int addCard(ushort caid, int provid, int sa, int ua)
        {
            this.cardCount++;
            IntPtr pCard = feynman_add_local_card(this.pServer, this.cardCount);
            feynman_set_local_card_caid(pCard, caid);
            feynman_set_local_card_provid(pCard, provid);
            feynman_set_local_card_sa(pCard, sa);
            feynman_set_local_card_ua(pCard, ua);

            return this.cardCount;
        }

        public unsafe void setSidRule(IntPtr card, ushort sid, bool allow)
        {
            char sidRule;
            if(allow)
            {
                sidRule = (char)1;
            }
            else
            {
                sidRule = (char)0;
            }
            feynman_add_sid(card, sid, sidRule);
        }

        public unsafe void start()
        {
            feynman_start_local_server(this.pServer);
        }
    }
}
