/***************************************************************************
 *   Copyright (C) 2007 by Dr. Feynman                                     *
 *   dr@feynman.com                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef WIN32
#include <winsock2.h>
#include "feynman_win32.h"
#else
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#define closesocket(x) close(x)
#endif
#include <fcntl.h>
#include <errno.h>

#ifdef USING_OPENSSL
#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

#ifdef USING_CYASSL
#include <cyassl/crypto.h>
#include <cyassl/ssl.h>
#include <cyassl/err.h>
#endif

#ifdef HAVE_SSL
  #ifndef SSL_SUCCESS
    #define SSL_SUCCESS     1
  #endif
#endif

#ifndef ETIMEDOUT
#define ETIMEDOUT 116
#endif

#include "feynman.h"

#ifdef HAVE_SSL
int feynman_init_remote_server_ssl(t_feynman_remote_server *server, char *cert, char *key)
{
    SSL_library_init();
    SSL_load_error_strings();

    SSL_METHOD *meth = TLSv1_client_method();

    server->ctx = SSL_CTX_new(meth);
    if(!server->ctx)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_CTX_new() failed\n");
        return FEYNMAN_ERROR;
    }

    if(!SSL_CTX_load_verify_locations(server->ctx, cert, NULL))
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_CTX_load_verify_locations() failed\n");
        return FEYNMAN_ERROR;
    }

    SSL_CTX_set_verify(server->ctx, SSL_VERIFY_NONE, NULL);

    return feynman_init_remote_server(server);
}
#endif

#ifdef WIN32
void feynman_remote_cache_thread(void *arg)
#else
void *feynman_remote_cache_thread(void *arg)
#endif
{
    t_feynman_ecm *ecm;
    t_feynman_emm *emm;

    t_feynman_remote_server *server = (t_feynman_remote_server *)arg;

#ifdef WIN32
    if(server == NULL) return;
   // if(server->dcw_callback == NULL) return;
#else
    if(server == NULL) return NULL;
   // if(server->dcw_callback == NULL) return NULL;
#endif

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Cache cleaning thread started\n");

    while(1)
    {
        struct timeval time_diff;
        int count = 0;
        int del_count = 0;

        feynman_lock_remote_server(server);

        for(ecm=server->ecm_db; ecm!=NULL; ecm=ecm->next)
        {
            gettimeofday(&time_diff, NULL);
            feynman_get_time_diff(&ecm->time_added, &time_diff);

            if(time_diff.tv_sec > 30)
            {
                feynman_del_remote_ecm(server, ecm);
                del_count++;
            }
            count++;
        }
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Flushed %d outdated ecms (%d left in pool)\n", del_count, count-del_count);
        count = 0;
        del_count = 0;
        for(emm=server->emm_db; emm!=NULL; emm=emm->next)
        {
            gettimeofday(&time_diff, NULL);
            feynman_get_time_diff(&emm->time_added, &time_diff);

            //30 minutes
            if(time_diff.tv_sec > 1800)
            {
                feynman_del_remote_emm(server, emm);
                del_count++;
            }
            count++;
        }
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Flushed %d outdated emms (%d left in pool)\n", del_count, count-del_count);

        feynman_unlock_remote_server(server);

        //20 seconds
        usleep(1000*1000*30);
    }
#ifndef WIN32
    return NULL;
#endif
}

#ifdef WIN32
void feynman_remote_socket_thread(void *arg)
#else
void *feynman_remote_socket_thread(void *arg)
#endif
{
    fd_set read_fds;
    fd_set master;
    unsigned char buf[FEYNMAN_BUF_SIZE];
    int len;
    int ret;
    int fdmax;
    struct timeval tv;

    t_feynman_remote_server *server = (t_feynman_remote_server *)arg;

#ifdef WIN32
    if(server == NULL) return;
#else
    if(server == NULL) return NULL;
#endif

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Socket thread started\n");

    FD_ZERO(&master);

    for(;;)
    {
        if(!server->ready)
        {
            feynman_lock_remote_server(server);
            server->logged_in = 0;
            if(FEYNMAN_STATUS_OK != feynman_connect_remote_server(server))
            {
                feynman_unlock_remote_server(server);
                usleep(3000000);
                continue;
            }
            if(FEYNMAN_STATUS_OK != feynman_send_login(server))
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Login failed\n");
                feynman_close_remote_server(server);
                feynman_unlock_remote_server(server);
                usleep(3000000);
                continue;
            }
            FD_ZERO(&read_fds);
            FD_ZERO(&master);
            FD_SET(server->handle.fd, &master);
            feynman_unlock_remote_server(server);
        }
        tv.tv_sec = 1;
        tv.tv_usec = 0;
        fdmax = server->handle.fd;
        read_fds = master;
        ret = select(fdmax+1, &read_fds, NULL, NULL, &tv);
        if(ret < 0)
        {
            perror("[ Feynman ] select");
            continue;
        }
        else if(ret == 0)
        {
            continue;
        }


        feynman_lock_remote_server(server);
#ifdef USING_CYASSL
        int errnum = SSL_ERROR_NONE;
        do {
#endif
        len = feynman_read_network_message(server->handle, buf);
        if(len < 0)
        {
            feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Server went down\n");
            FD_CLR(server->handle.fd, &read_fds);
            feynman_del_remote_cards(server);
            feynman_close_remote_server(server);
            feynman_unlock_remote_server(server);
            continue;
        }
        else if(len > 0)
        {
            feynman_process_remote_server_msg(server, buf, len);
#ifdef USING_CYASSL
            errnum = SSL_get_error(server->handle.ssl, len);
#endif
        }
#ifdef USING_CYASSL
        } while(errnum == SSL_ERROR_NONE && len > 0);
#endif
        feynman_unlock_remote_server(server);

    }
#ifndef WIN32
    return NULL;
#endif
}

#ifdef HAVE_SSL
int SSL_connect_nb(SSL* ssl)
{
    int ret = SSL_connect(ssl);
    while (ret != SSL_SUCCESS && SSL_get_error(ssl, ret) == SSL_ERROR_WANT_READ)
    {
        usleep(100000);
        ret = SSL_connect(ssl);
    }
    if (ret != SSL_SUCCESS)
    {
        return -1;
    }
    return 0;
}
#endif

static int connect_nb(int fd, struct sockaddr_in *addr, struct timeval *tout)
{
    int ret;
    socklen_t len;
    fd_set rset;

#ifdef WIN32
    unsigned long flags = 1;
    ioctlsocket(fd, FIONBIO, &flags);
#else
    int flags = fcntl(fd, F_GETFL, 0);
    if(flags < 0) return -1;

    ret = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
    if(ret < 0) return -1;
#endif

    len = sizeof(struct sockaddr);
    ret = connect(fd, (struct sockaddr *)addr, len);
    if(ret < 0 && errno != EINPROGRESS)
    {
        ret = -errno;
        goto done;
    }

    if(ret == 0) goto done;

    /* now wait */
    FD_ZERO(&rset);
    FD_SET(fd, &rset);

    ret = select(fd + 1, NULL, &rset, NULL, tout);
    if(ret <= 0)
    {
        if (ret == 0) ret = -ETIMEDOUT;
        else ret = -errno;
        goto done;
    }

    if(FD_ISSET(fd, &rset))
    {
        int status;

        len = sizeof(ret);
        status = getsockopt(fd, SOL_SOCKET, SO_ERROR, (void *)&ret, &len);
        if (status < 0)
        {
            ret = -errno;
            goto done;
        }

        /* Oops - something wrong with connect */
        if(ret) ret = -ret;
    }

done:
//Block the socket again
#ifdef WIN32
 //   flags = 0;
 //   ioctlsocket(fd, FIONBIO, &flags);
#else
 //   fcntl(fd, F_SETFL, flags);
#endif
    return ret;
}

int feynman_connect_remote_server(t_feynman_remote_server *server)
{
    struct hostent *hostaddr;
    struct sockaddr_in socketAddr;
    struct timeval tout = {0, 0};

    if(server == NULL) return FEYNMAN_ERROR;

#ifdef WIN32
    WSADATA wsaData;
    WSAStartup (MAKEWORD (2, 2), &wsaData);
#endif

    hostaddr = gethostbyname(server->hostname);
    if(!hostaddr)
    {
        return FEYNMAN_ERROR;
    }
    if((server->handle.fd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
        return FEYNMAN_ERROR;
    }

    socketAddr.sin_family = AF_INET;
    socketAddr.sin_port = htons(server->port);
    socketAddr.sin_addr.s_addr = ((struct in_addr *) hostaddr->h_addr)->s_addr;

    tout.tv_sec = 5;

    if(connect_nb(server->handle.fd, &socketAddr, &tout) != 0)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Failed connecting to %s\n", server->hostname);
        feynman_close_remote_server(server);
        return FEYNMAN_ERROR;
    }

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Connected to %s\n", server->hostname);

#ifdef HAVE_SSL
    if(!server->ctx)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Warning: SSL not initialized\n");
        return FEYNMAN_ERROR;
    }

    if(server->handle.ssl)
    {
        SSL_free(server->handle.ssl);
    }

    server->handle.ssl = SSL_new(server->ctx);
    if(!server->handle.ssl)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_new() failed\n");
        feynman_close_remote_server(server);
        return FEYNMAN_ERROR;
    }

    SSL_set_fd(server->handle.ssl, server->handle.fd);

    if(SSL_connect_nb(server->handle.ssl) < 0)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_connect() failed\n");
        feynman_close_remote_server(server);
        return FEYNMAN_ERROR;
    }
#endif
    server->ready = 1;

    return FEYNMAN_STATUS_OK;
}

void feynman_close_remote_server(t_feynman_remote_server *server)
{
    if(server == NULL) return;
    if(server->handle.fd > 0)
    {
        closesocket(server->handle.fd);
        server->handle.fd = 0;
    }
#ifdef HAVE_SSL
    if(server->handle.ssl > 0)
    {
        SSL_shutdown(server->handle.ssl);
        SSL_free(server->handle.ssl);
        server->handle.ssl = NULL;
    }
#endif
    server->ready = 0;
    server->logged_in = 0;
}

//TODO: remove possible old cards
int feynman_init_remote_server(t_feynman_remote_server *server)
{
    if(!server) return FEYNMAN_ERROR;
    server->ready = 0;

#ifdef WIN32
    _beginthread(feynman_remote_socket_thread, 0, server);
    _beginthread(feynman_remote_cache_thread, 0, server);
#else
    pthread_create(&server->socket_thread, 0, feynman_remote_socket_thread, server);
    pthread_create(&server->ecm_thread, 0, feynman_remote_cache_thread, server);
#endif

    return FEYNMAN_STATUS_OK;
}

void feynman_lock_remote_server(t_feynman_remote_server *server)
{
    if(server == NULL) return;
#ifdef WIN32
    WaitForSingleObject(server->server_lock, INFINITE);
#else
    pthread_mutex_lock(&server->server_lock);
#endif
}

void feynman_unlock_remote_server(t_feynman_remote_server *server)
{
    if(server == NULL) return;
#ifdef WIN32
    ReleaseMutex(server->server_lock);
#else
    pthread_mutex_unlock(&server->server_lock);
#endif
}


t_feynman_remote_server *feynman_add_remote_server(t_feynman_remote_server *server_db)
{
    t_feynman_remote_server *server;

    server = (t_feynman_remote_server *)malloc(sizeof(t_feynman_remote_server));

    server->handle.fd = 0;
    server->handle.bytes_in = 0;
    server->handle.bytes_out = 0;
#ifdef HAVE_SSL
    server->handle.ssl = NULL;
#endif
    server->logged_in = 0;
    server->access_mode = 0;
    server->ready = 0;

    server->username = NULL;
    server->password = NULL;

    server->hostname = NULL;
    server->port = 0;

    server->card_db = NULL;
    server->ecm_db = NULL;
    server->emm_db = NULL;
    server->dcw_callback = NULL;
    server->dcw_callback_args = NULL;
    server->next = server_db;
#ifdef HAVE_SSL
    server->ctx = NULL;
#endif

#ifndef WIN32
    pthread_mutex_init(&server->server_lock, NULL);
#endif

    if(server_db != NULL)
    {
        server->id = server_db->id + 1;
    }
    else
    {
        server->id = 1;
    }

    server->ecm_db = NULL;

    return server;
}

t_feynman_remote_server *feynman_get_remote_server(t_feynman_remote_server *server_db, int server_id)
{
    t_feynman_remote_server *i = NULL;

    if(!server_db) return NULL;

    for(i=server_db; i!=NULL; i=i->next)
    {
        if(server_id == i->id)
        {
            return i;
        }
    }
    return NULL;
}


t_feynman_remote_server *feynman_del_remote_server(t_feynman_remote_server *server_db, int server_id)
{
    if(server_db == NULL) return NULL;

    t_feynman_remote_server *server = server_db;
    t_feynman_remote_server *prev = server_db;
    while(server != NULL)
    {
        t_feynman_remote_server *tmp_server = server->next;
        if(server_id == -1 || server_id == server->id)
        {
            if(prev != NULL)
            {
                prev->next = server->next;
            }
            if(server_db == server)
            {
                server_db = server->next;
            }
            feynman_del_remote_cards(server);
            feynman_del_remote_ecms(server);
            feynman_del_remote_emms(server);
            free(server->username);
            free(server->password);
            free(server->hostname);
#ifdef HAVE_SSL
            SSL_CTX_free(server->ctx);
#endif
            free(server);
            if(server_id != -1) return server_db;
        }
        prev = server;
        server = tmp_server;
    }
    return server_db;
}

void feynman_set_remote_server_username(t_feynman_remote_server *server, char *username)
{
    if(server == NULL) return;
    if(username == NULL) return;

    int len = strlen(username);
    server->username = (char *)malloc(len+1);
    memcpy(server->username, username, len);
    server->username[len] = '\0';
}

void feynman_set_remote_server_password(t_feynman_remote_server *server, char *password)
{
    if(server == NULL) return;
    if(password == NULL) return;

    int len = strlen(password);
    server->password = (char *)malloc(len+1);
    memcpy(server->password, password, len);
    server->password[len] = '\0';
}

void feynman_set_remote_server_hostname(t_feynman_remote_server *server, char *hostname)
{
    if(server == NULL) return;
    if(hostname == NULL) return;

    int len = strlen(hostname);
    server->hostname = (char *)malloc(len+1);
    memcpy(server->hostname, hostname, len);
    server->hostname[len] = '\0';
}

void feynman_set_remote_server_port(t_feynman_remote_server *server, int port)
{
    if(server == NULL) return;

    server->port = port;
}

void feynman_set_remote_server_dcw_callback(t_feynman_remote_server *server, t_feynman_dcw_callback callback)
{
    if(server == NULL) return;

    server->dcw_callback = callback;
}

void feynman_set_remote_server_dcw_callback_args(t_feynman_remote_server *server, void *args)
{
    if(server == NULL) return;

    server->dcw_callback_args = args;
}
